/*
 Alarm System Monitor - Door Sensor
 Version: 2.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 7/10/2014
 
 Door sensor which wirelessly connects to the alarm system server. It connects to the server every 8 seconds to determine 
 whether the door sensor needs to be turned on or off.
 
 */
 
// ATtiny24/44/84 Pin map
//
//               +-\/-+
//         VCC  1|o   |14  GND
//         PB0  2|    |13  PA0 Battery ADC
//     LED PB1  3|    |12  PA1 
//   Reset PB3  4|    |11  PA2 
//      CE PB2  5|    |10  PA3 Input from door sensor
//     CSN PA7  6|    |9   PA4 SCK
//      MI PA6  7|    |8   PA5 MO
//               +----+

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "setup.c"
#include "hmac-sha1/hmac-sha1.h"
#include "mirf.c"

// Current system state
uint8_t system_state = SYSTEM_OFF;
uint8_t system_state_received = SYSTEM_OFF;
uint8_t sensor_disabled = false;
volatile uint8_t magnet_moved = false;

// 256bit random number (you should change this)
uint8_t random_number[] = {0x27, 0x9B, 0x6D, 0xD8, 0xBA, 0x17, 0x62, 0xD7,
							0xBB, 0x14, 0xF7, 0xDF, 0xDA, 0x41, 0xEE, 0x66,
							0x8A, 0x7A, 0x8D, 0X6F, 0xFA, 0x4F, 0xDA, 0xFC,
							0x41, 0x01, 0x12, 0x3B, 0x9F, 0xF7, 0xF2, 0xAF};

// HMAC 256bit random key (you should change this)
uint8_t hmac_key[] = {0x01, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
							0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
							0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 
							0x02, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77};

// 128bit rolling code number (you should change this)
const uint8_t rolling_code_base[] PROGMEM = {0x25, 0xF9, 0x34, 0x5A, 0xE2, 0xFD, 0x8D, 0x00,
										0x4C, 0x9A, 0xF7, 0x7F, 0x69, 0xD1, 0x3F, 0x82};
uint8_t rolling_code[16] = {0};
uint16_t rolling_code_addition = 0;

uint8_t hmac_test[20]; // Used to test the HMAC received from the server
uint8_t data_in[24];
uint8_t data_out[24];

int main(void) {
	setup();
	
	while(1) {
		// Power down the nRF24L01
		POWERDOWN;
		
		// Sleep for 8 seconds
		watchdog_sleep(T8S);
		
		// Check battery voltage
		cbi(PCMSK0, PCINT3); // Remove interrupt to door pin so it won't interrupt us
		sbi(ADCSRA, ADEN); // Turn on ADC
		analog_read(batteryAdcPin); // First reading is ignored
		int batteryVoltage = analog_read(batteryAdcPin); 
		cbi(ADCSRA, ADEN); // Turn off ADC
		if (sensor_disabled == false) {
			sbi(PCMSK0, PCINT3); // Apply interrupt to door pin
			
			// Because we turned off the interrupt, lets check the door sensor output
			if (!(PINA & (1<<doorSensor))) { // Door sensor output is low
				magnet_moved = true;
			}
			else {
				magnet_moved = false;
			}
		}
		
		// If battery voltage below 2.5V then turn off the door sensor
		if (batteryVoltage <= lowVoltageDoor) {
			cbi(PCMSK0, PCINT3); // Remove interrupt to door pin
			system_state = SYSTEM_OFF;
			magnet_moved = false;
			sensor_disabled = true; // Disable sensor permanently
		}
		
		// If movement is detected, send an alert to the server to turn on the alarm
		if ((system_state == SYSTEM_ON || system_state == SYSTEM_TEST) && magnet_moved == true && sensor_disabled == false) {
			for (uint8_t x = 0; x < 24; x++) {
				data_out[x] = ALARM_TRIGGERED;
			}
			
			// Generate the new rolling code to use and generate the SHA1 block (160 bit) from our rolling code
			rolling_code_init(rolling_code_addition);
			sha1(data_out, rolling_code, 128); // SHA1 of our rolling code
			
			// Power up the nRF24L01
			TX_POWERUP;
			_delay_ms(3);
			
			// Transmit a few times to ensure the server gets it
			for (uint8_t x = 0; x < 10; x++) {
				mirf_transmit_data();
				_delay_ms(50);
			}
			
			// Power down the nRF24L01
			POWERDOWN;
		}
		
		// Increment our 256bit random number
		random_number[31]++;
		if (random_number[31] == 0) {
			for (int8_t a = 31; a >= 0; a--) {
				if (random_number[a] == 0) {
					random_number[(a-1)]++;
				}
				else {
					break;
				}
			}
		}
		
		// Generate the random block (160 bit) using SHA1 from our random number
		sha1(data_out, random_number, 256);
		
		// Set data_out to say we are a door sensor and our sensor number
		data_out[SENSOR_TYPE] = DOOR_SENSOR;
		data_out[SENSOR_NUMBER] = SENSOR_NO;
		data_out[BATTERY_VOLTAGE_HI] = batteryVoltage >> 8;
		data_out[BATTERY_VOLTAGE_LO] = batteryVoltage & 0xFF;
		
		// Check in the with server
		if (check_in()) {
			// Perform alarm state action if alarm state received is different to what we have set
			if (system_state != system_state_received) {
				system_state = system_state_received;
				
				if (system_state == SYSTEM_ON || system_state == SYSTEM_TEST) {
					sbi(PCMSK0, PCINT3); // Enable interrupt of door sensor pin
				}
				else {
					cbi(PCMSK0, PCINT0); // Remove interrupt of door sensor pin
				}
			}
		}
	}
}

ISR(WDT_vect) { }

ISR(ADC_vect) { }

// Pin change
ISR(PCINT0_vect) {
	if (PINA & (1<<doorSensor)) { // Door sensor is next to magnet
		magnet_moved = false;
	}
	else {
		magnet_moved = true;
		
		// Blink LED
		PORTB |= (1<<LED);
		_delay_ms(100);
		PORTB &= ~(1<<LED);
	}
}
