/*
 Alarm System Monitor - Remote
 Version: 2.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 7/10/2014
 
 Used to switch the alarm system to on, off or test mode.
 
 */
 
// ATtiny24/44/84 Pin map
//
//               +-\/-+
//         VCC  1|o   |14  GND
//         PB0  2|    |13  PA0 System on button
//         PB1  3|    |12  PA1 System off button
//   Reset PB3  4|    |11  PA2 System test button
//      CE PB2  5|    |10  PA3
//     CSN PA7  6|    |9   PA4 SCK
//      MI PA6  7|    |8   PA5 MO
//               +----+

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "setup.c"
#include "hmac-sha1/hmac-sha1.h"
#include "mirf.c"

uint8_t remote_control_state = SYSTEM_ON; // For turning alarm system on or off

// 256bit random number (you should change this when re-programming, each device should be different)
uint8_t random_number[] = {0x27, 0x9B, 0x6D, 0xD8, 0xBA, 0x17, 0x62, 0xD7,
									0xBB, 0x14, 0xF7, 0xDF, 0xDA, 0x41, 0xEE, 0x66,
									0x8A, 0x7A, 0x8D, 0X6F, 0xFA, 0x4F, 0xDA, 0xFC,
									0x41, 0x01, 0x12, 0x3B, 0x9F, 0xF7, 0xF2, 0xAF};

// HMAC 256bit random key (you should change this, each device should have the same key)
uint8_t hmac_key[] = {0x01, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
							0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
							0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 
							0x02, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77};

// 128bit rolling code for remote control (you should change this, each device should be different)
const uint8_t rc_code_base[] PROGMEM = {0xD8, 0xBF, 0xE9, 0x55, 0x9E, 0x32, 0xA7, 0x69, 
												0x01, 0xEB, 0x8B, 0x62, 0xB7, 0x50, 0x41, 0x9F};
uint8_t rc_code[16];
uint32_t rc_code_addition = 0;

uint8_t hmac_test[20]; // Used to test the HMAC received from the server
uint8_t data_in[24];
uint8_t data_out[24];

int main(void) {
	setup();
	
	while(1) {
		// Sleep
		sbi(GIMSK,PCIE0); // Enable pin change interrupt
		system_sleep();
		cbi(GIMSK,PCIE0); // Disable pin change interrupt
		
		_delay_ms(10); // Wait for button to settle
		
		// Read button state
		remote_control_state = NO_INPUT;
		uint8_t waitForInputCounter = 0;
		while (remote_control_state == NO_INPUT || waitForInputCounter < 20) {
			if (PINA & (1<<PA0)) {
				remote_control_state = SYSTEM_ON;
			}
			else if (PINA & (1<<PA1)) {
				remote_control_state = SYSTEM_OFF;
			}
			else if (PINA & (1<<PA2)) {
				remote_control_state = SYSTEM_TEST;
			}
			_delay_ms(5);
			waitForInputCounter++;
		}
		
		if (remote_control_state != NO_INPUT) {
			PORTB |= (1<<LED);// Light LED
			
			// Increment our 256bit random number
			random_number[31]++;
			if (random_number[31] == 0) {
				for (int8_t a = 31; a >= 0; a--) {
					if (random_number[a] == 0) {
						random_number[(a-1)]++;
					}
					else {
						break;
					}
				}
			}
			
			// Generate the random block (160 bit) using SHA1 from our random number
			sha1(data_out, random_number, 256);
			
			// Set data_out to say we are a remote control and that we want the latest rolling code
			data_out[SENSOR_TYPE] = REMOTE_CONTROL;
			data_out[REMOTE_REQUEST] = REMOTE_ROLLING_CODE;
			
			// Check in the with server
			if (check_in()) {
				// Clear all data
				for (uint8_t x = 0; x < 24; x++) {
					data_out[x] = 0;
				}
				
				// On or off or test?
				data_out[SENSOR_TYPE] = REMOTE_CONTROL;
				data_out[REMOTE_REQUEST] = remote_control_state;
				
				// Generate the new rc code to use and generate the SHA1 block (160 bit) from our rc code
				rc_code_init(rc_code_addition);
				sha1(data_out, rc_code, 128); // SHA1 of our rolling code
				
				// Debug - output our current rolling code addition
				//data_out[ROLLING_CODE_32BIT_1B] = rc_code_addition >> 24;
				//data_out[ROLLING_CODE_32BIT_2B] = rc_code_addition >> 16;
				//data_out[ROLLING_CODE_32BIT_3B] = rc_code_addition >> 8;
				//data_out[ROLLING_CODE_32BIT_4B] = rc_code_addition & 0xFF;
				
				// Transmit a few times to ensure the server gets it
				for (uint8_t x = 0; x < 5; x++) {
					mirf_transmit_data();
					_delay_ms(50);
				}
			}
			
			// Power down the nRF24L01
			POWERDOWN; 
			PORTB &= ~(1<<LED);
		}
	}
}

ISR(WDT_vect) { }

ISR(PCINT0_vect) { }
