/*
 Alarm System Monitor - Server
 Version: 2.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 7/10/2014
 
 The server listens for incoming requests from the sensors and tells them whether they should switch themselves on or off.
 If a sensor sends the alarm triggered command, we give priority to the sirens and remote controls and ignore the other sensors.
 Turning the alarm system on or off is done by remote control.
 
 avrdude -p atmega328p -c usbtiny -U lfuse:w:0xff:m -U hfuse:w:0xd9:m -U efuse:w:0xff:m 
 
 */

// ATmega328 Pin Map
// 
//                           +-\/-+
//               Reset PC6  1|o   |28  PC5
//                     PD0  2|    |27  PC4 Check in LED - Failed
//            Nokia TX PD1  3|    |26  PC3 Check in LED - Ok
//                     PD2  4|    |25  PC2 Shift Register Latch
//                     PD3  5|    |24  PC1 Shift Register Clock
//                     PD4  6|    |23  PC0 Shift Register Data
//                     VCC  7|    |22  GND
//                     GND  8|    |21  AREF
//                XTAL PB6  9|    |20  AVCC
//                XTAL PB7 10|    |19  PB5 SCK
//                     PD5 11|    |18  PB4 MO
//                     PD6 12|    |17  PB3 MI
//        Alarm on LED PD7 13|    |16  PB2 CSN
//  Siren Mosfet / LED PB0 14|    |15  PB1 CE
//                           +----+

#define F_CPU 12000000 // 12 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <string.h>
#include "hmac-sha1/hmac-sha1.h"
#include "setup.c"
#include "mirf.c"
#include "sms.c"

volatile uint8_t system_state = SYSTEM_OFF; // Current system state for the device, on, off or test mode
volatile uint8_t siren_state = SIREN_OFF; // Is the alarm siren on or off
volatile uint8_t check_in_state = false;
volatile uint8_t sensor_data[2][2];
volatile uint8_t leds_reset = 0;

// HMAC 256bit random key (you should change this)
uint8_t hmac_key[] = {0x01, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
							0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
							0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 
							0x02, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77};

// 128bit rolling code for sensors (you should change this)
uint8_t sensor_code[] = {0x25, 0xF9, 0x34, 0x5A, 0xE2, 0xFD, 0x8D, 0x00,
									0x4C, 0x9A, 0xF7, 0x7F, 0x69, 0xD1, 0x3F, 0x82,};
uint16_t sensor_code_addition = 0;

// 128bit rolling code for remote control (you should change this)
uint8_t rc_code[] = {0xD8, 0xBF, 0xE9, 0x55, 0x9E, 0x32, 0xA7, 0x69, 
							0x01, 0xEB, 0x8B, 0x62, 0xB7, 0x50, 0x41, 0x9F};
uint32_t rc_code_addition = 0;

uint8_t data_in[24];
uint8_t data_out[24];

int main(void) {
	setup();
	
	while(1) {
		RX_POWERUP;
		mirf_flush_rx_tx(); // Flush RX and TX FIFO
		mirf_CE_hi; // Start listening
		
		// Wait for incoming requests
		while (!(mirf_status() & (1<<RX_DR))) { 
			_delay_us(100);
		}
		mirf_CE_lo; // Stop listening
		
		// Process the request
		if (!process_check_in()) {
			check_in_state = fail;
		}
		else {
			check_in_state = pass;
		}
	}
}

ISR(WDT_vect) { }

// Timer0 overflow after 20ms, update check in LED
ISR(TIMER0_OVF_vect) {
	if (check_in_state == pass) {
		PORTC |= (1<<CHECK_IN_OK);
		PORTC &= ~(1<<CHECK_IN_FAIL);
		check_in_state = false;
	}
	else if (check_in_state == fail) {
		PORTC |= (1<<CHECK_IN_FAIL);
		PORTC &= ~(1<<CHECK_IN_OK);
		check_in_state = false;
	}
	else { // Clear both LEDs on next overflow
		PORTC &= ~(1<<CHECK_IN_OK);
		PORTC &= ~(1<<CHECK_IN_FAIL);
	}
}

// Timer1 overflow after 340ms, update sensor/system state LEDs
ISR(TIMER1_OVF_vect) {
	PORTC &= ~(1<<SHIFT_LATCH);
	shift_out(sensor_data[SHIFT_REGISTER_1][SENSORS]);
	shift_out(sensor_data[SHIFT_REGISTER_2][SENSORS]);
	PORTC |= (1<<SHIFT_LATCH);
	
	// Blink the LEDs which have low voltage
	for (uint8_t x = 0; x < 8; x++) {
		if (sensor_data[SHIFT_REGISTER_1][BATTERY_VOLTAGES] & (1<<x)) {
			sensor_data[SHIFT_REGISTER_1][SENSORS] ^= (1<<x);
		}
		if (sensor_data[SHIFT_REGISTER_2][BATTERY_VOLTAGES] & (1<<x)) {
			sensor_data[SHIFT_REGISTER_2][SENSORS] ^= (1<<x);
		}
	}
	
	// Clear LEDs after ~1.5 minutes
	leds_reset++;
	if (leds_reset == 255) {
		sensor_data[SHIFT_REGISTER_1][SENSORS] = 0;
		sensor_data[SHIFT_REGISTER_1][BATTERY_VOLTAGES] = 0;
		sensor_data[SHIFT_REGISTER_2][SENSORS] = 0;
		sensor_data[SHIFT_REGISTER_2][BATTERY_VOLTAGES] = 0;
	}
	
	if (system_state == SYSTEM_ON || system_state == SYSTEM_TEST) {
		PORTD |= (1<<SYSTEM_ON_LED);
	}
	else {
		PORTD &= ~(1<<SYSTEM_ON_LED);
	}
}
