/*
 Alarm System Monitor - Server
 Version: 2.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 7/10/2014
 
 */
 
/* ******** Original copyright notice from mirf.c below ********
	
	Copyright (c) 2007 Stefan Engelke <mbox@stefanengelke.de>
	
	Permission is hereby granted, free of charge, to any person 
	obtaining a copy of this software and associated documentation 
	files (the "Software"), to deal in the Software without 
	restriction, including without limitation the rights to use, copy, 
	modify, merge, publish, distribute, sublicense, and/or sell copies 
	of the Software, and to permit persons to whom the Software is 
	furnished to do so, subject to the following conditions:
	
	The above copyright notice and this permission notice shall be 
	included in all copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
	DEALINGS IN THE SOFTWARE.
	
	$Id$
 */

#include "mirf.h"
#include "nRF24L01.h"
#include <avr/io.h>
#include <avr/interrupt.h>

// Defines for setting the MiRF registers for transmitting or receiving mode
#define TX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )
#define POWERDOWN mirf_config_register(CONFIG, mirf_CONFIG | ( (0<<PWR_UP) ) )

extern volatile uint8_t system_state;
extern volatile uint8_t siren_state;
extern uint8_t sensor_code[];
extern uint16_t sensor_code_addition;
extern uint8_t rc_code[];
extern uint32_t rc_code_addition;
extern uint8_t hmac_key[];
extern uint8_t data_in[];
extern uint8_t data_out[];
extern volatile uint8_t sensor_data[2][2];

extern void SendSMS(const char *, const uint8_t *);
extern uint8_t RecipientA[];
extern uint8_t RecipientM[];

// Initialises pins to communicate with the MiRF module
void mirf_init(void) {
	// Define CSN and CE as output and set them to default
	DDRB |= (1<<CE);
	DDRB |= (1<<CSN);
	mirf_CE_lo;
	mirf_CSN_hi;
}

// Sets the important registers in the MiRF module and powers the module in receiving mode
void mirf_config(void) {
	// Set RF channel
	mirf_config_register(RF_CH, mirf_CH);
	
	// Set length of incoming payload 
	mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
	mirf_config_register(RX_PW_P1, mirf_PAYLOAD);
	
	// Set auto retransmit delay to 500uS and leave the auto retransmit count to 3
	mirf_config_register(SETUP_RETR, (1<<4) | 3);
	
	// Set RADDR and TADDR
	mirf_write_register(RX_ADDR_P0, TADDR, 5);
	mirf_write_register(RX_ADDR_P1, RADDR, 5);
	mirf_write_register(TX_ADDR, TADDR, 5);
	
	// Enable RX_ADDR_P0 and RX_ADDR_P1 address matching (auto acknowledge is enabled by default)
	mirf_config_register(EN_RXADDR, (1<<ERX_P0 | 1<<ERX_P1));
	
	// Power up in receiver mode
	RX_POWERUP;
	_delay_ms(3);
}

// Flush RX and TX FIFO
void mirf_flush_rx_tx(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_RX); // Flush RX
	mirf_CSN_hi; // Pull up chip select

	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_TX);  // Write cmd to flush tx fifo
	mirf_CSN_hi; // Pull up chip select
}

// Read the status register
uint8_t mirf_status(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status;
}

// Checks if data is available for reading
uint8_t mirf_data_ready(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<RX_DR);
}

// Checks if max retransmissions has been reached
uint8_t mirf_max_rt_reached(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<MAX_RT);
}

// Reads mirf_PAYLOAD bytes into data array
void mirf_get_data(uint8_t *data) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
	spi_read_data(data, mirf_PAYLOAD); // Read payload
	mirf_CSN_hi; // Pull up chip select
	mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
}

// Write one byte into the MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_transfer(value);
	mirf_CSN_hi;
}

// Reads an array of bytes from the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(R_REGISTER | (REGISTER_MASK & reg));
	spi_read_data(value, len);
	mirf_CSN_hi;
}

// Writes an array of bytes into the MiRF register
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_write_data(value, len);
	mirf_CSN_hi;
}

// Send data and wait for an ACK
uint8_t mirf_transmit_data(void) {
	TX_POWERUP; // Power up to transmitter mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	
	mirf_CSN_lo;  // Pull down chip select
	spi_transfer(W_TX_PAYLOAD); // Write cmd to write payload
	spi_write_data(data_out, mirf_PAYLOAD); // Write payload
	mirf_CSN_hi; // Pull up chip select
	
	mirf_CE_hi; // Start transmission
	_delay_us(15);
	mirf_CE_lo;
	
	uint8_t waitcount = 0;
	while (!(mirf_status() & (1<<TX_DS))) { // Wait until we receive an ACK back
		_delay_us(500); // Wait the auto retransmit time
		if (mirf_status() & (1<<MAX_RT)) { // If maximum retries reached
			mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
			mirf_CE_hi;
			_delay_us(15);
			mirf_CE_lo;
		}
		waitcount++;
		if (waitcount >= 50) { // If we haven't heard back in 25ms, exit
			mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
			return 0;
		}
	}
	
	mirf_config_register(STATUS,(1<<TX_DS)); // Reset status register
	
	return 1;
}

// Receive data
uint8_t mirf_receive_data(void) {
	RX_POWERUP; // Power up to receiver mode
	mirf_CE_hi; // Start listening
	
	for (uint8_t x = 0; x < 24; x++) {
		data_in[x] = 0; // Clear the data_in
	}
	
	uint8_t waitcount = 0;
	while (!(mirf_status() & (1<<RX_DR))) { // Wait until data is ready to be read
		_delay_us(500);
		waitcount++;
		if (waitcount >= 50) { // If we haven't heard back in 25ms, exit
			mirf_CE_lo; // Stop listening
			return 0;
		}
	}
	mirf_CE_lo; // Stop listening
	
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
	spi_read_data(data_in, mirf_PAYLOAD); // Read payload
	mirf_CSN_hi; // Pull up chip select
	mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
	
	return 1;
}

// Process the check in
uint8_t process_check_in(void) {
	if (!mirf_receive_data()) { return 0; }
	
	// Update sensor number on LEDs
	if (data_in[SENSOR_TYPE] != REMOTE_CONTROL && data_in[SYSTEM_STATE] != ALARM_TRIGGERED) {
		uint8_t shiftRegToUpdate = 0;
		if (data_in[SENSOR_NUMBER] >= 0 && data_in[SENSOR_NUMBER] <= 7) {
			shiftRegToUpdate = SHIFT_REGISTER_1;
		}
		else if (data_in[SENSOR_NUMBER] >= 8 && data_in[SENSOR_NUMBER] <= 15) {
			shiftRegToUpdate = SHIFT_REGISTER_2;
			data_in[SENSOR_NUMBER] = data_in[SENSOR_NUMBER] - 8;
		}
		
		sensor_data[shiftRegToUpdate][SENSORS] |= (1<<data_in[SENSOR_NUMBER]);
		
		// Check if battery voltage is low
		int batteryVoltage = ((int) data_in[BATTERY_VOLTAGE_HI] << 8) | data_in[BATTERY_VOLTAGE_LO];
		if (data_in[SENSOR_TYPE] == PIR_SENSOR) {
			if (batteryVoltage <= lowVoltagePIR) {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] |= (1<<data_in[SENSOR_NUMBER]);
			}
			else {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] &= ~(1<<data_in[SENSOR_NUMBER]);
			}
		}
		else if (data_in[SENSOR_TYPE] == SIREN_SENSOR) {
			if (batteryVoltage <= lowVoltageSiren) {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] |= (1<<data_in[SENSOR_NUMBER]);
			}
			else {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] &= ~(1<<data_in[SENSOR_NUMBER]);
			}
		}
		else if (data_in[SENSOR_TYPE] == DOOR_SENSOR) {
			if (batteryVoltage <= lowVoltageDoor) {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] |= (1<<data_in[SENSOR_NUMBER]);
			}
			else {
				sensor_data[shiftRegToUpdate][BATTERY_VOLTAGES] &= ~(1<<data_in[SENSOR_NUMBER]);
			}
		}
	}
	
	// Check if the request came from a PIR or Door sensor only if the siren isn't on
	if (siren_state != SIREN_ON && (data_in[SENSOR_TYPE] == PIR_SENSOR || data_in[SENSOR_TYPE] == DOOR_SENSOR)) {
		// Add the system state to data_in
		data_in[SYSTEM_STATE] = system_state;
		
		// Rolling number addition
		data_in[ROLLING_CODE_16BIT_HI] = sensor_code_addition >> 8;
		data_in[ROLLING_CODE_16BIT_LO] = sensor_code_addition & 0xFF;
		
		// Use HMAC_SHA1 on the random number received (160bit) + sensor type (8bit) + system state (8bit) + addition on rolling code (16bit) using the 256bit HMAC key
		hmac_sha1(data_out, hmac_key, 256, data_in, 192);
		
		// Add the sensor type back in and also the system state to the data_out so the client can use it after it's verified the HMAC signing of the random key
		data_out[SENSOR_TYPE] = data_in[SENSOR_TYPE];
		data_out[SYSTEM_STATE] = system_state;
		data_out[ROLLING_CODE_16BIT_HI] = data_in[ROLLING_CODE_16BIT_HI];
		data_out[ROLLING_CODE_16BIT_LO] = data_in[ROLLING_CODE_16BIT_LO];
		
		if (!mirf_transmit_data()) { return 0; }
	}
	
	// Siren request
	if (data_in[SENSOR_TYPE] == SIREN_SENSOR) {
		// Add the siren state to data_in
		data_in[SIREN_STATE] = siren_state;
		
		// Use HMAC_SHA1 on the random number received (160bit) + sensor type (8bit) + siren state (8bit) using the 256bit HMAC key
		hmac_sha1(data_out, hmac_key, 256, data_in, 176);
		
		// Add the sensor type back in and also the siren state to the data_out so the client can use it after it's verified the HMAC signing of the random key
		data_out[SENSOR_TYPE] = data_in[SENSOR_TYPE];
		data_out[SIREN_STATE] = siren_state;
		
		if (!mirf_transmit_data()) { return 0; }
	}
	
	// Remote control requests
	if (data_in[SENSOR_TYPE] == REMOTE_CONTROL) {
		// Asking for the latest rolling code
		if (data_in[REMOTE_REQUEST] == REMOTE_ROLLING_CODE) {
			// Rolling number addition
			data_in[ROLLING_CODE_32BIT_1B] = rc_code_addition >> 24;
			data_in[ROLLING_CODE_32BIT_2B] = rc_code_addition >> 16;
			data_in[ROLLING_CODE_32BIT_3B] = rc_code_addition >> 8;
			data_in[ROLLING_CODE_32BIT_4B] = rc_code_addition & 0xFF;
			
			// Use HMAC_SHA1 on the random number received (160bit) + rolling code addition (32bit) using the 256bit HMAC key
			hmac_sha1(data_out, hmac_key, 256, data_in, 192);
			
			// Add the rolling number addition back in
			data_out[ROLLING_CODE_32BIT_1B] = data_in[ROLLING_CODE_32BIT_1B];
			data_out[ROLLING_CODE_32BIT_2B] = data_in[ROLLING_CODE_32BIT_2B];
			data_out[ROLLING_CODE_32BIT_3B] = data_in[ROLLING_CODE_32BIT_3B];
			data_out[ROLLING_CODE_32BIT_4B] = data_in[ROLLING_CODE_32BIT_4B];
			
			if (!mirf_transmit_data()) { return 0; }
		}
		else { // Verify the rolling code and perform the action
			// SHA1 of our rolling code
			sha1(data_out, rc_code, 128);
			
			// Verify SHA1 (160 bit) to the client's SHA1
			if (memcmp(data_in, data_out, 20) != 0) {
				return 0;
			}
			
			// Increment our 128bit rolling code number
			rc_code[15]++;
			if (rc_code[15] == 0) {
				for (uint8_t a = 15; a >= 0; a--) {
					if (rc_code[a] == 0) {
						rc_code[(a-1)]++;
					}
					else {
						break;
					}
				}
			}
			
			// Increment the rolling code addition to tell remotes which random number to use
			rc_code_addition++;
			
			// Turn alarm system on, off or test mode
			system_state = data_in[REMOTE_REQUEST];
			
			// Sound siren for little to confirm action received
			PORTB |= (1<<SIREN_LOCAL);
			_delay_ms(50);
			PORTB &= ~(1<<SIREN_LOCAL);
			
			// If system is now off, turn off sirens
			if (system_state == SYSTEM_OFF) {
				siren_state = SIREN_OFF;
				PORTB &= ~(1<<SIREN_LOCAL);
				
				// Send SMS off
				SendSMS("Off", RecipientA);
				_delay_ms(500);
				SendSMS("Off", RecipientM);
			}
			else if (system_state == SYSTEM_TEST) { // In test mode, turn sirens on test mode too
				siren_state = SYSTEM_TEST;
				
				// Send SMS test
				SendSMS("Test", RecipientA);
				_delay_ms(500);
				SendSMS("Test", RecipientM);
			}
		}
	}
	
	// Sensor sending alarm triggered command, only process if siren isn't on already
	if (siren_state != SIREN_ON && data_in[SYSTEM_STATE] == ALARM_TRIGGERED) {
		// SHA1 of our rolling code
		sha1(data_out, sensor_code, 128);
		
		// Verify SHA1 (160 bit) to the client's SHA1
		if (memcmp(data_in, data_out, 20) != 0) {
			return 0;
		}
		
		// Increment our 128bit rolling code number
		sensor_code[15]++;
		if (sensor_code[15] == 0) {
			for (uint8_t a = 15; a >= 0; a--) {
				if (sensor_code[a] == 0) {
					sensor_code[(a-1)]++;
				}
				else {
					break;
				}
			}
		}
		
		// Increment the rolling code addition to tell sensors which random number to use
		sensor_code_addition++;
		
		// If alarm system is on, turn on sirens and send SMS
		if (system_state == SYSTEM_ON) {
			siren_state = SIREN_ON; 
			PORTB |= (1<<SIREN_LOCAL);
			
			// Send SMS alarm triggered
			SendSMS("On", RecipientA);
			_delay_ms(500);
			SendSMS("On", RecipientM);
		}
	}
	
	return 1;
}
