/*
 Alarm System Monitor - Siren
 Version: 2.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 15/08/2012
 Last Modified: 7/10/2014
 
 The siren contacts the server every second to determine whether it should sound it's alarm.
 If the alarm is on and the server hasn't responded to our message and it's been 10 minutes then we turn the alarm off.
 
 */
 
// ATtiny24/44/84 Pin map
//
//               +-\/-+
//         VCC  1|o   |14  GND
//         PB0  2|    |13  PA0 Battery ADC
//         PB1  3|    |12  PA1
//   Reset PB3  4|    |11  PA2 
//      CE PB2  5|    |10  PA3 Mosfet to turn on Siren
//     CSN PA7  6|    |9   PA4 SCK
//      MI PA6  7|    |8   PA5 MO
//               +----+

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <string.h>
#include "setup.c"
#include "hmac-sha1/hmac-sha1.h"
#include "mirf.c"

// Current siren state
volatile uint8_t siren_state = SIREN_OFF;
int ten_minute_counter = 0;

// 256bit random number (you should change this)
uint8_t random_number[] = {0x27, 0x9B, 0x6D, 0xD8, 0xBA, 0x17, 0x62, 0xD7,
							0xBB, 0x14, 0xF7, 0xDF, 0xDA, 0x41, 0xEE, 0x66,
							0x8A, 0x7A, 0x8D, 0X6F, 0xFA, 0x4F, 0xDA, 0xFC,
							0x41, 0x01, 0x12, 0x3B, 0x9F, 0xF7, 0xF2, 0xAF};

// HMAC 256bit random key (you should change this)
uint8_t hmac_key[] = {0x01, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77,
							0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
							0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 
							0x02, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77};

uint8_t hmac_test[20]; // Used to test the HMAC received from the server
uint8_t data_in[24];
uint8_t data_out[24];

int main(void) {
	setup();
	
	while(1) {
		// Sleep for 1 second
		watchdog_sleep(T1S);
		
		// Check battery voltage
		sbi(ADCSRA, ADEN); // Turn on ADC
		analog_read(batteryAdcPin); // First reading is ignored
		int batteryVoltage = analog_read(batteryAdcPin); 
		cbi(ADCSRA, ADEN); // Turn off ADC
		
		// Switch off the siren after ~10 minutes if we don't hear anything back from the server to turn the siren off
		// 600 seconds / (1 second watchdog + 50ms random block)
		if (ten_minute_counter >= 571) {
			PORTA &= ~(1<<Siren); // Turn off siren
			ten_minute_counter = 0;
		}
		ten_minute_counter++;
		
		// Increment our 256bit random number
		random_number[31]++;
		if (random_number[31] == 0) {
			for (int8_t a = 31; a >= 0; a--) {
				if (random_number[a] == 0) {
					random_number[(a-1)]++;
				}
				else {
					break;
				}
			}
		}
		
		// Generate the random block (160 bit) using SHA1 from our random number
		sha1(data_out, random_number, 256);
		
		// Set data_out to say we are a siren and our sensor number
		data_out[SENSOR_TYPE] = SIREN_SENSOR;
		data_out[SENSOR_NUMBER] = SENSOR_NO;
		data_out[BATTERY_VOLTAGE_HI] = batteryVoltage >> 8;
		data_out[BATTERY_VOLTAGE_LO] = batteryVoltage & 0xFF;
		
		// Check in the with server
		if (check_in()) {
			// Perform alarm siren action
			if (siren_state == SIREN_ON) {
				PORTA |= (1<<Siren); // Turn on siren
				ten_minute_counter = 0;
				watchdog_sleep(T2S); // Sleep for 2 seconds so other sirens can check in
			}
			else if (siren_state == SIREN_TEST) {
				PORTA |= (1<<Siren); // Turn on siren
				_delay_ms(50);
				PORTA &= ~(1<<Siren); // Turn off siren
				watchdog_sleep(T8S); // Sleep for 8 seconds so we don't set this siren on again for a while
			}
			else {
				PORTA &= ~(1<<Siren); // Turn off siren
				ten_minute_counter = 0;
			}
		}
		
		// Power down the nRF24L01
		POWERDOWN;
	}
}

ISR(WDT_vect) { }

ISR(ADC_vect) { }
